﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace TSP
{
    class TreeNode
    {
        public Dictionary<int, TreeNode> successors;
        public int site, leafsCount;
        public double evaluation;
        public TreeNode parrent;
        public bool isRoot = false;

        public TSPInput input;

        public int ComputeLeafs()
        {
            if (this.IsLeaf)
            {
                this.leafsCount = 1;
                return this.leafsCount;
            }
            int sum = 0;
            foreach (var item in successors.Values)
            {
                sum += item.ComputeLeafs();
            }
            this.leafsCount = sum;
            return sum;
        }

        public double GetMinimum()
        {
            if (IsLeaf)
                return evaluation;
            double min = double.MaxValue;
            foreach (var item in successors.Values)
            {
                double val = item.GetMinimum();
                if (val < min)
                    min = val;
            }
            return min;
        }

        public double GetMaximum()
        {
            if (IsLeaf)
                return evaluation;
            double max = double.MinValue;
            foreach (var item in successors.Values)
            {
                double val = item.GetMaximum();
                if (val > max)
                    max = val;
            }
            return max;
        }

        private TreeNode addSuccessor(int site)
        {
            TreeNode n = new TreeNode(this);
            n.site = site;
            this.successors.Add(site, n);
            return n;
        }

        private double getEvaluation()
        {
            if (!IsLeaf)
                return 0;
            double sum = input.GetDistance(site, 0);
            TreeNode pom = this;
            while (pom.parrent != null)
            {
                sum += input.GetDistance(pom.site, pom.parrent.site);
                pom = pom.parrent;
            }
            return sum;
        }

        public void CreateSuccessors(HashSet<int> used, int start)
        {
            if (used.Count >= input.NodesCount)
            {
                this.evaluation = getEvaluation();
                return;
            }
            for (int i = 0; i < input.NodesCount; i++)
            {
                if (used.Contains(i))
                    continue;
                if (used.Count == input.NodesCount - 1 && i <= start)
                    continue;
                used.Add(i);
                if (start < 0)
                    this.addSuccessor(i).CreateSuccessors(used, i);
                else
                    this.addSuccessor(i).CreateSuccessors(used, start);
                if (this.successors[i].successors.Count == 0 && used.Count < input.NodesCount)
                    this.removeSuccessor(i);

                used.Remove(i);
            }
        }

        public void CreateSuccessors(HashSet<int> used, int start, List<int> permutation)
        {
            if (used.Count >= input.NodesCount)
            {
                this.evaluation = getEvaluation();
                return;
            }
            foreach (var i in permutation)
            {
                if (used.Contains(i))
                    continue;
                if (used.Count == input.NodesCount - 1 && i <= start)
                    continue;
                used.Add(i);
                if (start < 0)
                    this.addSuccessor(i).CreateSuccessors(used, i, permutation);
                else
                    this.addSuccessor(i).CreateSuccessors(used, start, permutation);
                if (this.successors[i].successors.Count == 0 && used.Count < input.NodesCount)
                    this.removeSuccessor(i);

                used.Remove(i);
            }
        }

        private void removeSuccessor(int i)
        {
            this.successors.Remove(i);
        }

        public TreeNode(TSPInput input)
        {
            this.input = input;
            this.successors = new Dictionary<int, TreeNode>();
        }

        public TreeNode(TreeNode treeNode)
        {
            this.input = treeNode.input;
            this.successors = new Dictionary<int, TreeNode>();
            this.parrent = treeNode;

        }

        public int GetDepth()
        {
            if (this.successors.Count == 0)
                return 1;
            foreach (var item in successors.Values)
            {
                return item.GetDepth() + 1;
            }
            return 0;
        }

        public bool IsLeaf
        {
            get
            {
                return this.successors.Count == 0;
            }
        }
    }

    class TreeVisualizer
    {
        private PictureBox treePB, fitnessPB;
        private Graphics gTree, gFitness;
        private int width, height;
        public FitnessForm solver;
        private PointF previous;

        public void Draw(TreeNode node, double max, double min)
        {
            int depth = node.GetDepth();
            gTree.Clear(Color.White);
            gFitness.Clear(Color.White);
            drawNodes(node, 0, width, 0, (height / 1.5f) / depth, PointF.Empty);
            previous = PointF.Empty;
            drawValues(node, 0, width, max, min);

            treePB.Refresh();
            fitnessPB.Refresh();
        }

        private float getTransformedValue(double val, double min, double max)
        {
            return (float)((max - val) * (fitnessPB.Height - 20) / (max - min) + 10);
        }

        private void drawValues(TreeNode node, float leftBound, float rightBound, double maxVal, double minVal)
        {
            if (node.IsLeaf)
            {
                float v = getTransformedValue(node.evaluation, minVal, maxVal);
                gFitness.DrawLine(Pens.Green, leftBound, v, rightBound, v);
                if (!previous.IsEmpty)
                    gFitness.DrawLine(Pens.Blue, new PointF((rightBound + leftBound) / 2, v), previous);
                previous = new PointF((rightBound + leftBound) / 2, v);
                return;
            }
            float start = leftBound;
            foreach (var item in node.successors.Values)
            {
                float newSize = item.leafsCount * (rightBound - leftBound) / node.leafsCount;
                drawValues(item, start, start + newSize, maxVal, minVal);
                start += newSize;
            }
        }

        private void drawNodes(TreeNode node, float leftBound, float rightBound, float upperBound, float horizontalStep, PointF predecessor)
        {
            PointF loc = new PointF((leftBound + rightBound) / 2, upperBound);
            //g.DrawRectangle(Pens.Black, loc.X, loc.Y, nodeSize, nodeSize);

            gTree.DrawString(node.site.ToString(), new Font("Arial", 10), Brushes.Blue, loc);
            if (node.IsLeaf)
            {
                gTree.DrawString(((int)Math.Floor(node.evaluation)).ToString(), new Font("Arial", 10), Brushes.Blue, loc.X, loc.Y + 20);//loc.X/width * (height/2));
            }

            if (!predecessor.IsEmpty)
                gTree.DrawLine(Pens.Black, predecessor, loc);

            float start = leftBound;
            foreach (var item in node.successors.Values)
            {
                float newSize = item.leafsCount * (rightBound - leftBound) / node.leafsCount;
                drawNodes(item, start, start + newSize, upperBound + horizontalStep, horizontalStep, loc);
                start += newSize;
            }
        }

        public TreeVisualizer(FitnessForm s)
        {
            this.solver = s;
            this.treePB = s.treePictureBox;
            this.fitnessPB = s.fitnessPictureBox;

            this.width = treePB.Width;
            this.height = treePB.Height;
            this.treePB.Image = new Bitmap(width, height);
            gTree = Graphics.FromImage(treePB.Image);

            this.fitnessPB.Image = new Bitmap(fitnessPB.Width, fitnessPB.Height);
            gFitness = Graphics.FromImage(fitnessPB.Image);

        }
    }

    class TreeEvaluator
    {
        double previousValue = -1,
            distanceSum = 0;

        int minimaCount = 0;
        bool isRising = false;

        public double Eval(TreeNode root, List<int> permutation = null, bool isDistance = true)
        {
            minimaCount = 0;
            isRising = false;
            previousValue = -1;
            distanceSum = 0;

            if (permutation == null || permutation.Count == 0)
                traverse(root);
            else
                traverse(root, permutation);

            if (isDistance) return distanceSum;
            else return minimaCount;
        }

        private void traverse(TreeNode node)
        {
            if (node.IsLeaf)
            {
                bool newDirectionIsRising = node.evaluation > previousValue;
                if (previousValue >= 0)
                {
                    distanceSum += Math.Abs(previousValue - node.evaluation);
                    if (!isRising && newDirectionIsRising)
                        minimaCount++;
                }
                isRising = newDirectionIsRising;
                previousValue = node.evaluation;
            }
            foreach (var item in node.successors.Values)
            {
                traverse(item);
            }
        }

        private void traverse(TreeNode node, List<int> permutation)
        {
            if (node.IsLeaf)
            {
                bool newDirectionIsRising = node.evaluation > previousValue;
                if (previousValue >= 0)
                {
                    distanceSum += Math.Abs(previousValue - node.evaluation);
                    if (!isRising && newDirectionIsRising)
                        minimaCount++;
                }
                isRising = newDirectionIsRising;
                previousValue = node.evaluation;
            }
            foreach (var item in permutation)
            {
                if (node.successors.ContainsKey(item))
                    traverse(node.successors[item], permutation);
            }
        }
    }

}
