﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using LocalSearch;

namespace TSP
{
    struct TPSCouple
    {
        public int first, second;

        public TPSCouple(int first, int second)
        {
            this.first = first;
            this.second = second;
        }

        public int GetTheOther(int value)
        {
            if (first == value)
                return second;
            else return first;
        }

        public void Set(int value)
        {
            if (first == value)
                second = value;
            else first = value;
        }
    }

    class TSPSolution
    {
        private bool isDistanceComputed = false;
        private double _totalDistance;

        public TSPInput problem;

        private Dictionary<int, TPSCouple> succesors; //successors in the path. position of the neibhours indexed by the position of the node

        public double TotalDistance
        {
            get
            {
                if (!isDistanceComputed)
                    ComputeDistance();
                return _totalDistance;
            }
            private set
            {
                _totalDistance = value;
            }
        }
        public bool IsValid
        {
            get;
            private set;
        }
        public List<Edge> edges;

        public TSPSolution(TSPInput inp)
        {
            this.problem = inp;
            this.succesors = new Dictionary<int, TPSCouple>();
            this.edges = new List<Edge>();
        }
   
        public double ComputeDistance()
        {
            if (isDistanceComputed)
                return _totalDistance;
            _totalDistance = 0;
            for (int i = 0, j = 0; i < problem.NodesCount; i++, j = GetSuccessor(j))
            {
                _totalDistance += problem.GetDistance(j, GetSuccessor(j));
            }
            isDistanceComputed = true;
            return TotalDistance;
        }
        public int GetSuccessor(int node, int notWantedSuccesor)
        {
            return succesors[node].GetTheOther(notWantedSuccesor);
        }
        public int GetSuccessor(int node)
        {
            return succesors[node].first;
        }
        public void SetSuccessor(int node, int successor)
        {
            isDistanceComputed = false;
            if (!succesors.ContainsKey(node))
                succesors.Add(node, new TPSCouple(successor, successor));
            succesors[node].Set(successor);
            edges.Add(new Edge(node, successor, this.problem, problem.GetDistance(node, successor)));
        }
        public bool Validate()
        {
            IsValid = false;
            HashSet<int> visited = new HashSet<int>();
            int node = 0;
            while (!visited.Contains(node))
            {
                visited.Add(node);
                if (!succesors.ContainsKey(node))
                    return false;
                node = GetSuccessor(node);
            }
            if (visited.Count < problem.NodesCount)
                return false;

            IsValid = true;
            return true;
        }

        internal static TSPSolution fromString(string result, TSPInput inp)
        {
            string[] points = result.Split(new string[] { " " }, StringSplitOptions.RemoveEmptyEntries);
            if (points.Length != inp.NodesCount)
            {
                System.Windows.Forms.MessageBox.Show("Spatna delka reseni.\nReseni ma mit delku " + inp.NodesCount + ". Vase reseni ma delku " + points.Length);
                return null;
            }
            TSPSolution sol = new TSPSolution(inp);
            for (int i = 0; i < points.Length - 1; i++)
            {
                sol.SetSuccessor(int.Parse(points[i]) - 1, int.Parse(points[i + 1]) - 1);
            }
            sol.SetSuccessor(int.Parse(points[points.Length - 1]) - 1, int.Parse(points[0]) - 1);
            return sol;
        }
    }

    class PermutationStandard : IntegerSequenceCandidateSolution<TSPInput>
    {
        public int Size;

        public TSPSolution ConvertToTSPSol()
        {
            TSPSolution sol = new TSPSolution(problem);
            for (int i = 0; i < Size - 1; i++)
            {
                sol.SetSuccessor(data[i], data[i + 1]);
            }
            sol.SetSuccessor(data[Size - 1], data[0]);
            return sol;
        }

        public PermutationInversions ConvertToInversions()
        {
            PermutationInversions inv = new PermutationInversions(this.problem)
            {
                Size = this.Size,
                data = new int[this.Size]
            };
            for (int i = 0; i < this.Size; i++)
            {
                int count = 0, m = 0;
                while (this.data[m] != i)
                {
                    if (this.data[m] > i)
                        count++;
                    m++;
                }
                inv.data[i] = count;
            }
            return inv;
        }

        public PermutationStandard(TSPInput input)
            :base(new int[input.NodesCount], input)
        {
            this.Size = input.NodesCount;
        }

        public PermutationStandard(TSPSolution sol) : this(sol.problem)
        {
            data[0] = 0;
            data[1] = sol.GetSuccessor(data[0]);
            for (int i = 1; i < Size - 1; i++)
            {
                data[i + 1] = sol.GetSuccessor(data[i], data[i - 1]);
            }
        }

        public override double Evaluate()
        {
            double totalDistance = 0;
            for (int i = 0; i < Size - 1; i++)
            {
                totalDistance += problem.GetDistance(data[i], data[i + 1]);
            }
            totalDistance += problem.GetDistance(data[Size - 1], data[0]);
            return totalDistance;
        }

        public PermutationStandard Clone()
        {
            PermutationStandard res = new PermutationStandard(this.problem);
            res.Size = this.Size;
            res.data = new int[this.Size];
            Array.Copy(this.data, res.data, this.Size);
            return res;
        }

        public override IntegerSequenceCandidateSolutionFactory<IntegerSequenceCandidateSolution<TSPInput>, TSPInput> GetFactory()
        {
            return PermutationStandardFactory.GetInstance();
        }
    }

    class PermutationInversions : IntegerSequenceCandidateSolution<TSPInput>
    {
        public int Size;

        public PermutationStandard ConvertToStandard()
        {
            PermutationStandard std = new PermutationStandard(this.problem);
            int[] pos = new int[this.Size];
            for (int i = Size - 1; i >= 0; i--)
            {
                for (int m = i; m < Size; m++)
                {
                    if (pos[m] >= this.data[i] + 1)
                        pos[m]++;
                    pos[i] = data[i] + 1;
                }
            }
            for (int i = 0; i < Size; i++)
            {
                std.data[pos[i] - 1] = i;
            }
            return std;
        }

        public TSPSolution ConvertToTSPSol()
        {
            return this.ConvertToStandard().ConvertToTSPSol();
        }

        public PermutationInversions(TSPInput input)
            : base(new int[input.NodesCount], input)
        {
            this.Size = input.NodesCount;
        }

        public override double Evaluate()
        {
            return ConvertToStandard().Evaluate();
            //return convertToTSPSol().totalDistance;
        }

        public override IntegerSequenceCandidateSolutionFactory<IntegerSequenceCandidateSolution<TSPInput>, TSPInput> GetFactory()
        {
            return PermutationInversionsFactory.GetInstance();
        }
    }

    class PermutationStandardFactory : IntegerSequenceCandidateSolutionFactory<PermutationStandard, TSPInput>
    {
        public PermutationStandard Create(int[] data, TSPInput problem)
        {
            var res = new PermutationStandard(problem);
            res.data = data;
            return res;
        }

        public PermutationStandard CreateEmpty(TSPInput problem)
        {
            var res = new PermutationStandard(problem);
            res.data = new int[problem.NodesCount];
            return res;
        }

        public static PermutationStandardFactory GetInstance()
        {
            return singleton.Value;
        }

        private static Lazy<PermutationStandardFactory> singleton = new Lazy<PermutationStandardFactory>(() => new PermutationStandardFactory());

        private PermutationStandardFactory()
        { }
    }

    class PermutationInversionsFactory : IntegerSequenceCandidateSolutionFactory<PermutationInversions, TSPInput>
    {
        public PermutationInversions Create(int[] data, TSPInput problem)
        {
            var res = new PermutationInversions(problem);
            res.data = data;
            return res;
        }

        public static PermutationInversionsFactory GetInstance()
        {
            return singleton.Value;
        }

        private static Lazy<PermutationInversionsFactory> singleton = new Lazy<PermutationInversionsFactory>(() => new PermutationInversionsFactory());

        private PermutationInversionsFactory()
        {  }

    }

    class PermutationRandomizer : LSNullaryOperator<PermutationStandard, TSPInput>
    {
        Random r;

        public PermutationStandard Apply(TSPInput Problem)
        {
            var res = PermutationStandardFactory.GetInstance().CreateEmpty(Problem);
            int size = Problem.NodesCount;

            List<int> notUsed = new List<int>();
            for (int i = 0; i < size; i++)
            {
                notUsed.Add(i);
            }
            for (int i = 0; i < size; i++)
            {
                int selectedIndex = r.Next(notUsed.Count);
                res.data[i] = notUsed[selectedIndex];
                notUsed.RemoveAt(selectedIndex);
            }
            return res;
        }

        public PermutationRandomizer(Random r)
        {
            this.r = r;
        }
    }

    class PermutationRandomizer2 : LSNullaryOperator<PermutationStandard, TSPInput>
    {
        Random r;

        public PermutationRandomizer2(Random r)
        {
            this.r = r;
        }

        public PermutationStandard Apply(TSPInput problem)
        {
            var res = PermutationStandardFactory.GetInstance().CreateEmpty(problem);
            int size = problem.NodesCount;

            int distanceused = 0, allUsed = 0;
            List<int> notUsed = new List<int>();
            for (int i = 1; i < size; i++)
            {
                notUsed.Add(i);
            }
            res.data[0] = 0;

            double distanceSum = 0;
            int count = 0;
            for (int i = 0; i < size; i++)
            {
                for (int j = i + 1; j < size; j++)
                {
                    distanceSum += problem.GetDistance(i, j);
                    count++;
                }
            }
            double distanceLimit = (distanceSum / count) / 2;

            for (int i = 1; i < size; i++)
            {
                var nearByCities = notUsed.Where(a => problem.GetDistance(res.data[i - 1], a) < distanceLimit);
                if (nearByCities.Count() == 0)
                {
                    allUsed++;
                    int selectedIndex = r.Next(notUsed.Count);
                    res.data[i] = notUsed[selectedIndex];
                    notUsed.RemoveAt(selectedIndex);
                }
                else
                {
                    distanceused++;
                    int selectedIndex = r.Next(nearByCities.Count());
                    res.data[i] = nearByCities.ToList()[selectedIndex];
                    List<int> ff = nearByCities.ToList();
                    notUsed.Remove(res.data[i]);
                }
            }
            return res;
        }
    }
}
