﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace TSP
{
    class TreeAproximationSolver : TSPSolver
    {
        public Dictionary<int, List<int>> successorsList;
        List<Edge> edges;
        int[] representatives;
        int lastNodeLabeled;
        HashSet<int> visited = new HashSet<int>();
        
        private int findRepresentative(int node)
        {
            int representative = node;
            while (representatives[representative] != representative)
                representative = representatives[representative];
            representatives[node] = representative;
            return representative;
        }

        private void addEdge(Edge e)
        {
            if (!successorsList.ContainsKey(e.From))
                successorsList.Add(e.From, new List<int>());
            if (!successorsList.ContainsKey(e.To))
                successorsList.Add(e.To, new List<int>());
            successorsList[e.From].Add(e.To);
            successorsList[e.To].Add(e.From);
        }

        private void computeSpanningTree(TSPInput input)
        {
            representatives = new int[input.NodesCount];
            for (int i = 0; i < input.NodesCount; i++)
            {
                representatives[i] = i;
            }
            successorsList = new Dictionary<int, List<int>>();
            copyEdgesLengths(input);
            int selectedEdgesCount = 0;
            for (int i = 0; i < edges.Count; i++)
            {
                if (selectedEdgesCount == input.NodesCount - 1)
                    break;
                int firstRepre = findRepresentative(edges[i].From),
                    secondRepre = findRepresentative(edges[i].To);
                if (firstRepre == secondRepre)
                    continue;
                representatives[firstRepre] = secondRepre;
                selectedEdgesCount++;
                addEdge(edges[i]);
            }
        }

        private void labelTreePreOrder(int start, TSPSolution solution)
        {
            if (visited.Contains(start))
                return;
            lastNodeLabeled = start;
            visited.Add(start);
            foreach (var item in successorsList[start])
            {
                if (visited.Contains(item))
                    continue;
                solution.SetSuccessor(lastNodeLabeled, item);
                labelTreePreOrder(item, solution);
            }
        }

        private void copyEdgesLengths(TSPInput input)
        {
            edges = new List<Edge>();
            for (int i = 0; i < input.NodesCount; i++)
                for (int j = i + 1; j < input.NodesCount; j++)
                    edges.Add(new Edge(i, j, input, input.GetDistance(i, j)));
            edges.Sort((t1, t2) => t1.Weight.CompareTo(t2.Weight));
        }

        #region TSPSolver Members

        public TSPSolution Solve(TSPInput input)
        {
            computeSpanningTree(input);
            TSPSolution result = new TSPSolution(input);
            visited.Clear();
            labelTreePreOrder(0, result);
            result.SetSuccessor(lastNodeLabeled, 0);
            return result;
            //return null;
        }

        #endregion
    }

    class TreeApproximativeSolverFactory : TSPSolverFactory<TreeAproximationSolver>
    {
        #region TSPSolverFactory<TreeAproximativeSolver> Members

        public TreeAproximationSolver Create()
        {
            return new TreeAproximationSolver();
        }

        #endregion
    }
}
